// RXMQCR.CPP : A DLL to give REXX Support for IBM/MQ/OS2 MQSC
//
// Written by, and only begetter, Robert Harris,
//                                MQ/OS2 Performance,
//                                IBM Hursley.
//
// This module builds into a DLL with provides Rexx Support for
//      MQ/OS2 MQSC. Note that the name of the resultant DLL is
//      set to the THISDLL macro. (I also use the trigraph
//      representation for square brackets, as this gets round
//      character set problems).
//
//
//
//
//   In order to use this Rexx/MQ Interface, this DLL's initialisation
//      function must be called before usage.
//
//   This is done by a RxFuncAdd('RXMQCRINIT','RXMQCR',RXMQCRINIT')
//        call to let Rexx know about the DLL, and then a
//        rcc = RXMQCRINIT() to call the initialisation function.
//        As part of this initialisation function, all the other
//        RXMQCR functions are registered with Rexx.
//
//   There is no equivalent termination function, as the routines
//         attach to process termination to stop access at this
//         time, and a Termination operation would interfere
//         with this processing. However, a function called
//         RXMQCRTERM will deregister the interface functions.
//
//   All the RXMQCR functions return a standard Rexx Return string.
//       This is structured so that the numeric Return Code
//       (which may be negative) is obtained by a word(RCC,1) call.
//
//   The Return Code for an operation can be negative to
//         show that this DLL has detected the error, otherwise
//         it will be the MQ Completion Code
//
//         The Return String is in text format as follows:
//
//            Word 1 : Return Code
//            Word 2 : MQ Completion Code (or 0 if MQ not done)
//            Word 2 : MQ Reason     Code (or 0 if MQ not done)
//            Word 4 : RXMQCR function being run
//            Word > : OK or an helpful error message
//
//
//   Tracing MAY be enabled (if this module has been compiled with
//           the relevant setting of the DEBUG macro) by setting
//           values into a Rexx Variable call RXMQCRTRACE. The
//           settings are:
//
//                              COM   -> Command Function
//
//                              GV    -> Obtaining a Rexx Variable
//                              PV    -> Setting   a Rexx Variable
//
//                              INIT  -> Initialisation processing
//                              TERM  -> Deregistration processing
//
//                              *     -> Trace everything!!!
//
//
//  Copyright : This program is Copyrighted
//              by IBM UK Labs Limited, 1995.
//
//
//   Support  : This program is not supported in any way by IBM.
//              It is distributed only to show techniques for
//              Message Queueing (if so distributed).
//              It may be freely modified and adapted by your
//              installation.
//
//   Author   : Robert Harris,
//              Message Queueing/Performance,
//              IBM Hursley Park,
//              England.
//
//
//
//

//
// Standard Header file includes
//

 #include <string.h>
 #include <stdlib.h>
 #include <stdio.h>
 #include <iostream.h>
 #include <time.h>

//
// OS/2 Header file includes
//

 #define INCL_DOSMEMMGR
 #include <os2.h>

//
// Rexx Header file includes
//

 #define INCL_RXSHV
 #define INCL_RXFUNC
 #include <REXXSAA.H>

//
// MQ/OS2 Header file includes
//

 #include <P:\MQ\BIN\CMQC.H>
 #include <P:\MQ\BIN\CMQCFC.H>
 #include <P:\MQ\BIN\CMQXC.H>


//
//  The DEBUG macro is used to enable tracing. This may be perminantly
//            enabled/disabled, or dynamically under the control
//            of the RXMQCRTRACE Rexx variable.
//
//
//  DEBUG settings :
//
//  DEBUG  -> istrace   for dynamic debugging based on RXMQLTRACE
//  DEBUG  -> cout <<   for static trace everything debugging
//  DEBUG  -> ;         for stop all tracing
//
//  DEBUGS -> cout <<   for    tracing in istrace (lots of lines!!!)
//  DEBUGS -> ;         for no tracing in istrace routine
//
//
//  The setting are thus:
//
// #define DEBUG(p1,p2) if (istrace( p1 ) == TRUE) cout << p2 << endl
// #define DEBUG(p1,p2) cout << p2 << endl
// #define DEBUG(p1,p2)
//
// #define DEBUGS(p1) cout << p1 << endl
// #define DEBUGS(p1)
//

#define DEBUG(p1,p2) if (istrace( p1 ) == TRUE) cout << p2 << endl
#define DEBUGS(p1)

//
// Global definitions
//
//  MAXQS      -> maximum number of Queues allowed to process
//  THISDLL    -> name of this DLL
//  MAXCOMMLEN -> maximum size of command permitted
//
//

#define THISDLL "RXMQCR"
#define MAXCOMMLEN 5000

//
// Definitions of datatypes
//

#define FULLWORD     long
#define HALFWORD     int
#define MAXSFULLWORD (68719476735)
#define MAXUFULLWORD (4294967295)
#define MAXSHALFWORD (4095)
#define MAXUHALFWORD (65535)
#define MAXFILENAME  (255)

//
// All the actual MQ calls are preformed within the RAHMQO
//     class library. This is bought in at compile time
//     for ease of customer compilation. There is no problem
//     about using the Linkage editor to bind at link time
//     if this is so desired.
//

//
// RAHMQO Class Library
//

class rahmqo
 {
   public :

    char        qm  ??(MQ_Q_MGR_NAME_LENGTH+1??) ; //QM name
    char        q   ??(MQ_Q_NAME_LENGTH+1    ??) ; //Q  name
    MQHCONN     qmh                              ; //QM handle
    MQHOBJ      qh                               ; //Q  handle
    MQLONG      rc                               ; //Return Code
    MQLONG      ac                               ; //Reason Code
    MQLONG      lastrc                           ; //Prev Return Code
    MQLONG      lastac                           ; //Prev Reason Code


    rahmqo ( void )               ;  //Contructor for new  use of QM
    rahmqo ( MQHCONN const uqmh)  ;  //Contructor for same use of QM

    MQLONG      conn ( void )               ;  //Connect    to QM
    MQLONG      conn ( char * const uqm )   ;  //  with QM name supplied

    MQLONG      disc ( void )               ;  //Disconnect from QM

    void        cleanup ( void )            ;  //Close and Disconnect

    MQOD        odopen                      ;  //OD for MQOPEN usage
    MQLONG      open (MQLONG const uopts )  ;  //MQOPEN call
    MQLONG      open (char * const uq    ,     //  with Q name supplied
                      MQLONG const uopts )  ;  //
    MQLONG      open (MQOD * const uod   ,     //  with OD and options
                      MQLONG const uopts )  ;  //

    MQLONG      close (MQLONG const uopts ) ;  //MQCLOSE call

    MQLONG      commit   ( void ) ; //Syncpoint UOW
    MQLONG      rollback ( void ) ; //Backout   UOW

    MQLONG      inq ( MQLONG   const  uselnum ,             //MQINQ call
                      MQLONG          usels   ,
                      MQLONG   const  uinum   ,
                      MQLONG *        uiattr  ,
                      MQLONG   const  ucharl  ,
                      char   *        ucattr  ) ;

    MQLONG      set ( MQLONG   const uselnum  ,             //MQSET call
                      MQLONG         usels    ,
                      MQLONG   const uinum    ,
                      MQLONG *       uiattr   ,
                      MQLONG   const ucharl   ,
                      char   *       ucattr   ) ;

    MQMD        pmd                             ;           //PUT Message Dwescriptor
    MQLONG      put ( MQPMO  * const  uopts   ,             //MQPUT call
                      MQLONG   const  bufferl ,
                      char   * const  buffer  ) ;

    MQLONG      put ( MQMD   *        upmd    ,
                      MQPMO  * const  uopts   ,
                      MQLONG   const  bufferl ,
                      char   * const  buffer  ) ;

                                                            //PUT Zero-Length record
    MQLONG      putz ( MQPMO  * const  uopts     ) ;        //No Tag, Use Instance MD

    MQLONG      putz ( MQMD   *        upmd      ,          //No Tag, use supplied MD
                       MQPMO  * const  uopts     ) ;

    MQLONG      putz ( MQPMO    * const  uopts   ,          //Tag, Use Instance MD
                       MQBYTE24 * const  utag    ) ;

    MQLONG      putz ( MQMD     *        upmd    ,          //Tag, use supplied MD
                       MQPMO    * const  uopts   ,
                       MQBYTE24 * const  utag    ) ;

    MQMD        bmd                                  ;      //Browse Message Descriptor
    MQGMO       bmo                                  ;      //Browse MQGET options
    MQLONG      browse ( MQLONG   const  bufferl   ,
                         char   * const  buffer    ,
                         MQLONG *        udatalen  ) ;

    MQLONG      browse ( MQMD   *        ubmd      ,
                         MQLONG   const  bufferl   ,
                         char   * const  buffer    ,
                         MQLONG *        udatalen  ) ;


    MQMD        gmd                               ;         //GET Message Descriptor
    MQLONG      get ( MQGMO  * const  uopts     ,           //MQGET call
                      MQLONG   const  bufferl   ,
                      char   * const  buffer    ,
                      MQLONG *        udatalen  ) ;

    MQLONG      get ( MQMD   *        ugmd      ,
                      MQGMO  * const  uopts     ,
                      MQLONG   const  bufferl   ,
                      char   * const  buffer    ,
                      MQLONG *        udatalen  ) ;


    double      gettimeconn     () { return (time_conn     ) ; }
    double      gettimedisc     () { return (time_disc     ) ; }
    double      gettimeopen     () { return (time_open     ) ; }
    double      gettimeclose    () { return (time_close    ) ; }
    double      gettimeinq      () { return (time_inq      ) ; }
    double      gettimeset      () { return (time_set      ) ; }
    double      gettimeput      () { return (time_put      ) ; }
    double      gettimeget      () { return (time_get      ) ; }
    double      gettimecommit   () { return (time_commit   ) ; }
    double      gettimerollback () { return (time_rollback ) ; }
    double      getcputconn     () { return (cput_conn     / CLOCKS_PER_SEC) ; }
    double      getcputdisc     () { return (cput_disc     / CLOCKS_PER_SEC) ; }
    double      getcputopen     () { return (cput_open     / CLOCKS_PER_SEC) ; }
    double      getcputclose    () { return (cput_close    / CLOCKS_PER_SEC) ; }
    double      getcputinq      () { return (cput_inq      / CLOCKS_PER_SEC) ; }
    double      getcputset      () { return (cput_set      / CLOCKS_PER_SEC) ; }
    double      getcputput      () { return (cput_put      / CLOCKS_PER_SEC) ; }
    double      getcputget      () { return (cput_get      / CLOCKS_PER_SEC) ; }
    double      getcputcommit   () { return (cput_commit   / CLOCKS_PER_SEC) ; }
    double      getcputrollback () { return (cput_rollback / CLOCKS_PER_SEC) ; }

 private :

    double      time_conn      ;
    double      time_disc      ;
    double      time_open      ;
    double      time_close     ;
    double      time_inq       ;
    double      time_set       ;
    double      time_put       ;
    double      time_get       ;
    double      time_commit    ;
    double      time_rollback  ;


    double      cput_conn      ;
    double      cput_disc      ;
    double      cput_open      ;
    double      cput_close     ;
    double      cput_inq       ;
    double      cput_set       ;
    double      cput_put       ;
    double      cput_get       ;
    double      cput_commit    ;
    double      cput_rollback  ;
 } ;

//
// rahmqo member function definitions
//

//
// Constructor functions
//
//     without a parm, then this is a new use of a QM
//     with          , then this is an additional access on a
//                     previously connected QM
//

rahmqo::rahmqo ( void )
{
  memset(qm,0,sizeof(qm)) ;
  memset(q ,0,sizeof(q )) ;
  qmh    = 0              ;
  qh     = 0              ;
  rc     = 0              ;
  ac     = 0              ;
  lastrc = 0              ;
  lastac = 0              ;
  memset(&odopen,0,sizeof(odopen))                             ;
  strncpy(odopen.StrucId,MQOD_STRUC_ID,sizeof(odopen.StrucId)) ;
  odopen.Version    = MQOD_VERSION_1                           ;
  odopen.ObjectType = MQOT_Q                                   ;
  memset(&pmd,0,sizeof(pmd))                                   ;
  memset(&bmd,0,sizeof(bmd))                                   ;
  memset(&gmd,0,sizeof(gmd))                                   ;

  cput_conn      = 0 ;
  cput_disc      = 0 ;
  cput_open      = 0 ;
  cput_close     = 0 ;
  cput_inq       = 0 ;
  cput_set       = 0 ;
  cput_put       = 0 ;
  cput_get       = 0 ;
  cput_commit    = 0 ;
  cput_rollback  = 0 ;
  time_conn      = 0 ;
  time_disc      = 0 ;
  time_open      = 0 ;
  time_close     = 0 ;
  time_inq       = 0 ;
  time_set       = 0 ;
  time_put       = 0 ;
  time_get       = 0 ;
  time_commit    = 0 ;
  time_rollback  = 0 ;

  return             ;
}

rahmqo::rahmqo ( MQHCONN const uqmh )
{
  memset(qm,0,sizeof(qm)) ;
  memset(q ,0,sizeof(q )) ;
  qmh    = uqmh           ;
  qh     = 0              ;
  rc     = 0              ;
  ac     = 0              ;
  lastrc = 0              ;
  lastac = 0              ;
  memset(&odopen,0,sizeof(odopen))                             ;
  strncpy(odopen.StrucId,MQOD_STRUC_ID,sizeof(odopen.StrucId)) ;
  odopen.Version    = MQOD_VERSION_1                           ;
  odopen.ObjectType = MQOT_Q                                   ;
  memset(&pmd,0,sizeof(pmd))                                   ;
  memset(&bmd,0,sizeof(bmd))                                   ;
  memset(&gmd,0,sizeof(gmd))                                   ;

  cput_conn      = 0 ;
  cput_disc      = 0 ;
  cput_open      = 0 ;
  cput_close     = 0 ;
  cput_inq       = 0 ;
  cput_set       = 0 ;
  cput_put       = 0 ;
  cput_get       = 0 ;
  cput_commit    = 0 ;
  cput_rollback  = 0 ;
  time_conn      = 0 ;
  time_disc      = 0 ;
  time_open      = 0 ;
  time_close     = 0 ;
  time_inq       = 0 ;
  time_set       = 0 ;
  time_put       = 0 ;
  time_get       = 0 ;
  time_commit    = 0 ;
  time_rollback  = 0 ;

  return             ;
}

//
// rahmqo destructor functions
//
//   these are not provided
//
//    (Note that a disc call removes access to ALL
//     instances of RAHMQO sharing the Handle. as I
//     am not aiming at a 'proper' C++ implementation
//     I have not provided code to cope with this event).
//

//
// rahmqo functions to access a Queue Manager
//

//
// conn  will connect to a Queue Manager
//       saving the QMH
//
//       The version with no parms will use the
//           pre-set Queue Manager name
//
//       The alternative access is to quote a
//           Queue Manager Name which is the
//           one to be used
//

MQLONG  rahmqo::conn ( char * const uqm )
{
  double c1, c2, t1, t2         ;
  strncpy(qm,uqm,sizeof(qm)-1)  ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQCONN ( qm, &qmh, &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_conn += 1.0000 * ( c2 - c1 )      ;
  time_conn += 1.0000 * difftime(t2,t1)  ;

  return (ac) ;
}


MQLONG  rahmqo::conn ( void )
{
  double c1, c2, t1, t2         ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQCONN ( qm, &qmh, &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_conn += 1.0000 * ( c2 - c1 )      ;
  time_conn += 1.0000 * difftime(t2,t1)  ;

  return (ac) ;
}


//
// disc  will disconnect from a Queue Manager
//       deleting the QMH
//


MQLONG  rahmqo::disc ( void )
{
  double c1, c2, t1, t2         ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQDISC ( &qmh, &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_disc += 1.0000 * ( c2 - c1 )      ;
  time_disc += 1.0000 * difftime(t2,t1)  ;

  qmh = 0     ;

  return (ac) ;
}

//
// cleanup will Close the Q (if known) and Disconnect from
//         the QM (if known). This will effect other instances
//         which were created with a QM handle
//

void    rahmqo::cleanup ( void )
{
  double c1, c2, t1, t2         ;

  if ( qh  != 0 ) close(MQCO_NONE) ;
  if ( qmh != 0 ) disc ()          ;
  return                           ;
}

//
// commit will Syncpoint ALL accesses to the
//       QM on the given QMH (this will effect
//       other instances of the Class if the
//       QMH is being shared) in the current UOW
//

MQLONG  rahmqo::commit ( void )
{
  double c1, c2, t1, t2         ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQCMIT ( qmh, &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_commit += 1.0000 * ( c2 - c1 )      ;
  time_commit += 1.0000 * difftime(t2,t1)  ;

  return (ac) ;
}

//
// rollback will backout ALL accesses to the
//       QM on the given QMH (this will effect
//       other instances of the Class if the
//       QMH is being shared) in the current UOW
//

MQLONG  rahmqo::rollback ( void )
{
  double c1, c2, t1, t2         ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQBACK ( qmh, &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_rollback += 1.0000 * ( c2 - c1 )      ;
  time_rollback += 1.0000 * difftime(t2,t1)  ;

  return (ac) ;
}

//
// open  will provide access to a Queue
//
//       The OPEN options are always supplied
//           by the caller.
//
//       If the version without a Queue Name is
//          used, then the Name from the Instance
//          is taken.
//

MQLONG  rahmqo::open ( MQLONG const uopts )
{
  double c1, c2, t1, t2         ;

  memset(&odopen,0,sizeof(odopen))                             ;
  strncpy(odopen.StrucId,MQOD_STRUC_ID,sizeof(odopen.StrucId)) ;
  odopen.Version    = MQOD_VERSION_1                           ;
  odopen.ObjectType = MQOT_Q                                   ;
  strcpy(odopen.ObjectName,q)                                  ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQOPEN ( qmh, &odopen, uopts, &qh, &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_open += 1.0000 * ( c2 - c1 )      ;
  time_open += 1.0000 * difftime(t2,t1)  ;

  return (ac) ;
}

MQLONG  rahmqo::open ( char * const uq    ,
                       MQLONG const uopts )
{
  double c1, c2, t1, t2         ;

  strncpy(q,uq,sizeof(q)-1)   ;

  memset(&odopen,0,sizeof(odopen))                             ;
  strncpy(odopen.StrucId,MQOD_STRUC_ID,sizeof(odopen.StrucId)) ;
  odopen.Version    = MQOD_VERSION_1                           ;
  odopen.ObjectType = MQOT_Q                                   ;
  strcpy(odopen.ObjectName,q)                                  ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQOPEN ( qmh, &odopen, uopts, &qh, &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_open += 1.0000 * ( c2 - c1 )      ;
  time_open += 1.0000 * difftime(t2,t1)  ;

  return (ac) ;
}

MQLONG  rahmqo::open ( MQOD * const uod   ,
                       MQLONG const uopts )
{
  double c1, c2, t1, t2         ;

  memcpy(&odopen,uod,sizeof(odopen)) ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQOPEN ( qmh, &odopen, uopts, &qh, &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_open += 1.0000 * ( c2 - c1 )      ;
  time_open += 1.0000 * difftime(t2,t1)  ;

  memcpy(uod,&odopen,sizeof(odopen))     ;
  return (ac) ;
}

//
// close will stop access to a Queue
//
//       The CLOSE options are always supplied
//           by the caller.
//

MQLONG  rahmqo::close ( MQLONG const uopts )
{
  double c1, c2, t1, t2         ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQCLOSE ( qmh, &qh, uopts, &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_close += 1.0000 * ( c2 - c1 )      ;
  time_close += 1.0000 * difftime(t2,t1)  ;

  qh = 0      ;

  return (ac) ;
}

//
// inq   will inquire upon the Attributes of
//       a Queue etc.
//
//       Note that ALL the Selectors stuff
//       has to be supplied by the caller
//

MQLONG  rahmqo::inq ( MQLONG   const  uselnum ,
                      MQLONG          usels   ,
                      MQLONG   const  uinum   ,
                      MQLONG *        uiattr  ,
                      MQLONG   const  ucharl  ,
                      char   *        ucattr  )
{
  double c1, c2, t1, t2         ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQINQ ( qmh, qh, uselnum, &usels ,
                   uinum  , uiattr ,
                   ucharl , ucattr ,
                                     &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_inq += 1.0000 * ( c2 - c1 )      ;
  time_inq += 1.0000 * difftime(t2,t1)  ;

  return (ac) ;
}
//
// set   will set the Attributes of a Queue etc.
//
//       Note that ALL the Selectors stuff
//       has to be supplied by the caller
//

MQLONG  rahmqo::set ( MQLONG   const   uselnum ,
                      MQLONG           usels   ,
                      MQLONG   const   uinum   ,
                      MQLONG *         uiattr  ,
                      MQLONG   const   ucharl  ,
                      char   *         ucattr  )
{
  double c1, c2, t1, t2         ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQSET ( qmh, qh, uselnum , &usels  ,
                   uinum   , uiattr  ,
                   ucharl  , ucattr  ,
                                       &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_set += 1.0000 * ( c2 - c1 )      ;
  time_set += 1.0000 * difftime(t2,t1)  ;

  return (ac) ;
}


//
// put       will write a Message into a Queue
//
//           The Message descriptor may either
//           be supplied or the instance version
//           used.
//

MQLONG  rahmqo::put ( MQPMO  * const  uopts   ,
                      MQLONG   const  bufferl ,
                      char   * const  buffer  )
{
  double c1, c2, t1, t2         ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQPUT ( qmh, qh, &pmd, uopts, bufferl, buffer, &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_put += 1.0000 * ( c2 - c1 )      ;
  time_put += 1.0000 * difftime(t2,t1)  ;

  return (ac) ;
}


MQLONG  rahmqo::put ( MQMD   *        upmd    ,
                      MQPMO  * const  uopts   ,
                      MQLONG   const  bufferl ,
                      char   * const  buffer  )
{
  double c1, c2, t1, t2         ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQPUT ( qmh, qh, upmd, uopts, bufferl, buffer, &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_put += 1.0000 * ( c2 - c1 )      ;
  time_put += 1.0000 * difftime(t2,t1)  ;

  return (ac) ;
}

//
// putz      will write a Zero-Length Message into a Queue
//
//           If a Tag is supplied, then this is used as the Message Descriptor
//              this must be an area of MQBYTE24 (24 bytes long)
//
//           The Message descriptor may either
//           be supplied or the instance version
//           used.
//

MQLONG  rahmqo::putz ( MQPMO  * const  uopts )               //No Tag, Use Instance MD
{
  double c1, c2, t1, t2         ;

  memset(&this->pmd.MsgId,0,sizeof(this->pmd.MsgId)) ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQPUT ( qmh, qh, &pmd, uopts, 0, 0, &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_put += 1.0000 * ( c2 - c1 )      ;
  time_put += 1.0000 * difftime(t2,t1)  ;

  return (ac) ;
}


MQLONG  rahmqo::putz ( MQMD   *        upmd    ,              //No Tag, use supplied MD
                       MQPMO  * const  uopts   )
{
  double c1, c2, t1, t2         ;

  memset(upmd->MsgId,0,sizeof(upmd->MsgId)) ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQPUT ( qmh, qh, upmd, uopts, 0, 0, &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_put += 1.0000 * ( c2 - c1 )      ;
  time_put += 1.0000 * difftime(t2,t1)  ;

  return (ac) ;
}

MQLONG  rahmqo::putz ( MQPMO    * const  uopts   ,            //Tag, Use Instance MD
                       MQBYTE24 * const  utag    )
{
  double c1, c2, t1, t2         ;

  memcpy(&this->pmd.MsgId,utag,sizeof(this->pmd.MsgId)) ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQPUT ( qmh, qh, &pmd, uopts, 0, 0, &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_put += 1.0000 * ( c2 - c1 )      ;
  time_put += 1.0000 * difftime(t2,t1)  ;

  return (ac) ;
}


MQLONG  rahmqo::putz ( MQMD     *        upmd    ,            //Tag, use supplied MD
                       MQPMO    * const  uopts   ,
                       MQBYTE24 * const  utag    )
{
  double c1, c2, t1, t2         ;

  memcpy(upmd->MsgId,utag,sizeof(upmd->MsgId)) ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQPUT ( qmh, qh, upmd, uopts, 0, 0, &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_put += 1.0000 * ( c2 - c1 )      ;
  time_put += 1.0000 * difftime(t2,t1)  ;

  return (ac) ;
}

//
//    Browse    will non-destructively read a Queue
//
//              Versions are supplied that use an
//              instance Message dewscriptor, or
//              one supplied by the caller
//

MQLONG  rahmqo::browse ( MQLONG   const  bufferl   ,
                         char   * const  buffer    ,
                         MQLONG *        udatalen  )
{
  double c1, c2, t1, t2         ;

  memset(&bmo,0,sizeof(bmo))                                 ;
  memcpy(&bmo.StrucId,MQGMO_STRUC_ID,sizeof(bmo.StrucId))    ;
  bmo.Version        = MQGMO_VERSION_1                       ;
  bmo.Options        = MQGMO_NO_WAIT+MQGMO_BROWSE_NEXT+MQGMO_ACCEPT_TRUNCATED_MSG+
                       MQGMO_FAIL_IF_QUIESCING               ;
  bmo.WaitInterval   = MQWI_UNLIMITED                        ;

  memset(&bmd,0,sizeof(bmd))                                 ;
  memcpy(&bmd.StrucId,MQMD_STRUC_ID,sizeof(bmd.StrucId))     ;
  bmd.Version        = MQMD_VERSION_1                        ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQGET ( qmh, qh, &bmd, &bmo, bufferl, buffer, udatalen, &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_get += 1.0000 * ( c2 - c1 )      ;
  time_get += 1.0000 * difftime(t2,t1)  ;

  return (ac) ;
}


MQLONG  rahmqo::browse ( MQMD   *        ubmd      ,
                         MQLONG   const  bufferl   ,
                         char   * const  buffer    ,
                         MQLONG *        udatalen  )
{
  double c1, c2, t1, t2         ;

  memset(&bmo,0,sizeof(bmo))                                 ;
  memcpy(&bmo.StrucId,MQGMO_STRUC_ID,sizeof(bmo.StrucId))    ;
  bmo.Version        = MQGMO_VERSION_1                       ;
  bmo.Options        = MQGMO_NO_WAIT+MQGMO_BROWSE_NEXT+MQGMO_ACCEPT_TRUNCATED_MSG+
                       MQGMO_CONVERT+MQGMO_FAIL_IF_QUIESCING ;
  bmo.WaitInterval   = MQWI_UNLIMITED                        ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQGET ( qmh, qh, ubmd, &bmo, bufferl, buffer, udatalen, &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_get += 1.0000 * ( c2 - c1 )      ;
  time_get += 1.0000 * difftime(t2,t1)  ;

  return (ac) ;
}


MQLONG  rahmqo::get ( MQGMO  * const  uopts     ,           //MQGET call
                      MQLONG   const  bufferl   ,
                      char   * const  buffer    ,
                      MQLONG *        udatalen  )
{
  double c1, c2, t1, t2         ;

  memset(&gmd,0,sizeof(gmd))                                 ;
  memcpy(&gmd.StrucId,MQMD_STRUC_ID,sizeof(gmd.StrucId))     ;
  bmo.Version        = MQMD_VERSION_1                        ;

  memset(&gmd,0,sizeof(gmd))                                 ;
  memcpy(&gmd.StrucId,MQMD_STRUC_ID,sizeof(gmd.StrucId))     ;
  gmd.Version        = MQMD_VERSION_1                        ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQGET ( qmh, qh, &gmd, uopts, bufferl, buffer, udatalen, &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_get += 1.0000 * ( c2 - c1 )      ;
  time_get += 1.0000 * difftime(t2,t1)  ;

  return (ac) ;
}


MQLONG  rahmqo::get ( MQMD   *        ugmd      ,
                      MQGMO  * const  uopts     ,
                      MQLONG   const  bufferl   ,
                      char   * const  buffer    ,
                      MQLONG *        udatalen  )
{
  double c1, c2, t1, t2         ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQGET ( qmh, qh, ugmd, uopts, bufferl, buffer, udatalen, &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_get += 1.0000 * ( c2 - c1 )      ;
  time_get += 1.0000 * difftime(t2,t1)  ;

  return (ac) ;
}

//
// End of RAHMQO Class definition and functions
//


//
// Global Variables
//


//
// Polymorphic C++ functions
//

//
// Output Stream functions - convert MQ objects into cout
//
//        The MQOD, MQMD, MQPMO and MQGMO structures are
//            enabled for use with the << output stream
//            operators via the usual operator overloadings.
//
//        Each field is dumped in character and hex.
//
//        (At the time of writing, then functions are
//         only enabled when the revelant trace is active)
//

//
// Print a MQOD  - Object descriptor
//

ostream& operator<<(ostream& os, MQOD * od)
 {

  LONG     i    ; //scanner

  os << "MQOD Structid        :"  ;
        for(i=0; i < sizeof(MQCHAR4) ;
            os << od->StrucId??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQCHAR4) ;
            os << hex << (unsigned short)od->StrucId??(i++??)) ;
  os << "> " ;
  os << endl ;

  os << "     Version         :"  ;
        os << od->Version ;
  os << " <"  ;
        os << hex << od->Version ;
  os << "> " ;
  os << endl ;

  os << "     ObjectType      :"  ;
        os << od->ObjectType ;
  os << " <"  ;
        os << hex << od->ObjectType ;
  os << "> " ;
  os << endl ;

  os << "     ObjectName      :"  ;
        for(i=0; i < sizeof(MQCHAR48) ;
            os << od->ObjectName??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQCHAR48) ;
            os << hex << (unsigned short)od->ObjectName??(i++??)) ;
  os << "> " ;
  os << endl ;

  os << "     ObjectQMgrName  :"  ;
        for(i=0; i < sizeof(MQCHAR48) ;
            os << od->ObjectQMgrName??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQCHAR48) ;
            os << hex << (unsigned short)od->ObjectQMgrName??(i++??)) ;
  os << "> " ;
  os << endl ;

  os << "     DynamicQName    :"  ;
        for(i=0; i < sizeof(MQCHAR48) ;
            os << od->DynamicQName??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQCHAR48) ;
            os << hex << (unsigned short)od->DynamicQName??(i++??)) ;
  os << "> " ;
  os << endl ;

  os << "     AlternateUserId :"  ;
        for(i=0; i < sizeof(MQCHAR12) ;
            os << od->AlternateUserId??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQCHAR12) ;
            os << hex << (unsigned short)od->AlternateUserId??(i++??)) ;
  os << "> " ;
  os << endl ;

  return (os) ;

 }

//
// Print a MQGMO - Get Message Options block
//

ostream& operator<<(ostream& os, MQGMO * gmo)
 {

  LONG     i    ; //scanner

  os << "MQGMO Structid        :"  ;
        for(i=0; i < sizeof(MQCHAR4) ;
            os << gmo->StrucId??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQCHAR4) ;
            os << hex << (unsigned short)gmo->StrucId??(i++??)) ;
  os << "> " ;
  os << endl ;

  os << "      Version         :"  ;
        os << gmo->Version ;
  os << " <"  ;
        os << hex << gmo->Version ;
  os << "> " ;
  os << endl ;

  os << "      Options         :"  ;
        os << gmo->Options   ;
  os << " <"  ;
        os << hex << gmo->Options ;
  os << "> " ;
  os << endl ;

  os << "      WaitInterval    :"  ;
        os << gmo->WaitInterval ;
  os << " <"  ;
        os << hex << gmo->WaitInterval ;
  os << "> " ;
  os << endl ;

  os << "      Signal1         :"  ;
        os << gmo->Signal1 ;
  os << " <"  ;
        os << hex << gmo->Signal1 ;
  os << "> " ;
  os << endl ;

  os << "      Signal2         :"  ;
        os << gmo->Signal2 ;
  os << " <"  ;
        os << hex << gmo->Signal2 ;
  os << "> " ;
  os << endl ;

  os << "      ResolvedQName   :"  ;
        for(i=0; i < sizeof(MQCHAR48) ;
            os << gmo->ResolvedQName??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQCHAR48) ;
            os << hex << (unsigned short)gmo->ResolvedQName??(i++??)) ;
  os << "> " ;
  os << endl ;

  return (os) ;

 }

//
// Print a MQPMO - Put Message Options block
//

ostream& operator<<(ostream& os, MQPMO * pmo )
 {

  LONG     i    ; //scanner

  os << "MQPMO Structid        :"  ;
        for(i=0; i < sizeof(MQCHAR4) ;
            os << pmo->StrucId??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQCHAR4) ;
            os << hex << (unsigned short)pmo->StrucId??(i++??)) ;
  os << "> " ;
  os << endl ;

  os << "      Version         :"  ;
        os << pmo->Version ;
  os << " <"  ;
        os << hex << pmo->Version ;
  os << "> " ;
  os << endl ;

  os << "      Options         :"  ;
        os << pmo->Options ;
  os << " <"  ;
        os << hex << pmo->Options ;
  os << "> " ;
  os << endl ;

  os << "      Timeout         :"  ;
        os << pmo->Timeout ;
  os << " <"  ;
        os << hex << pmo->Timeout ;
  os << "> " ;
  os << endl ;

  os << "      Context         :"  ;
        os << pmo->Context ;
  os << " <"  ;
        os << hex << pmo->Context ;
  os << "> " ;
  os << endl ;

  os << "      KnownDestCount  :"  ;
        os << pmo->KnownDestCount ;
  os << " <"  ;
        os << hex << pmo->KnownDestCount ;
  os << "> " ;
  os << endl ;

  os << "      UnknownDestCount:"  ;
        os << pmo->UnknownDestCount ;
  os << " <"  ;
        os << hex << pmo->UnknownDestCount ;
  os << "> " ;
  os << endl ;

  os << "      InvalidDestCount:"  ;
        os << pmo->InvalidDestCount ;
  os << " <"  ;
        os << hex << pmo->InvalidDestCount ;
  os << "> " ;
  os << endl ;

  os << "      ResolvedQName   :"  ;
        for(i=0; i < sizeof(MQCHAR48) ;
            os << pmo->ResolvedQName??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQCHAR48) ;
            os << hex << (unsigned short)pmo->ResolvedQName??(i++??)) ;
  os << "> " ;
  os << endl ;

  os << "      ResolvedQMgrName:"  ;
        for(i=0; i < sizeof(MQCHAR48) ;
            os << pmo->ResolvedQMgrName??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQCHAR48) ;
            os << hex << (unsigned short)pmo->ResolvedQMgrName??(i++??));
  os << "> " ;
  os << endl ;


  return (os) ;

 }

//
// Print a MQMD  - MQ Message Descriptor
//

ostream& operator<<(ostream& os, MQMD * md )
 {

  LONG     i    ; //scanner

  os << "MQMD  Structid        :"  ;
        for(i=0; i < sizeof(MQCHAR4) ;
            os << md->StrucId??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQCHAR4) ;
            os << hex << (unsigned short)md->StrucId??(i++??)) ;
  os << "> " ;
  os << endl ;

  os << "      Version         :"  ;
        os << md->Version ;
  os << " <"  ;
        os << hex << md->Version ;
  os << "> " ;
  os << endl ;

  os << "      Report          :"  ;
        os << md->Report ;
  os << " <"  ;
        os << hex << md->Report ;
  os << "> " ;
  os << endl ;

  os << "      MsgType         :"  ;
        os << md->MsgType ;
  os << " <"  ;
        os << hex << md->MsgType ;
  os << "> " ;
  os << endl ;

  os << "      Expiry          :"  ;
        os << md->Expiry  ;
  os << " <"  ;
        os << hex << md->Expiry ;
  os << "> " ;
  os << endl ;

  os << "      Feedback        :"  ;
        os << md->Feedback ;
  os << " <"  ;
        os << hex << md->Feedback ;
  os << "> " ;
  os << endl ;

  os << "      Encoding        :"  ;
        os << md->Encoding ;
  os << " <"  ;
        os << hex << md->Encoding ;
  os << "> " ;
  os << endl ;

  os << "      CodedCharSetId  :"  ;
        os << md->CodedCharSetId ;
  os << " <"  ;
        os << hex << md->CodedCharSetId ;
  os << "> " ;
  os << endl ;

  os << "      Format          :"  ;
        for(i=0; i < sizeof(MQCHAR8) ;
            os << md->Format??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQCHAR8) ;
            os << hex << (unsigned short)md->Format??(i++??)) ;
  os << "> " ;
  os << endl ;

  os << "      Priority        :"  ;
        os << md->Priority ;
  os << " <"  ;
        os << hex << md->Priority ;
  os << "> " ;
  os << endl ;

  os << "      Persistence     :"  ;
        os << md->Persistence ;
  os << " <"  ;
        os << hex << md->Persistence ;
  os << "> " ;
  os << endl ;

  os << "      MsgId           :"  ;
        for(i=0; i < sizeof(MQBYTE24) ;
            os << md->MsgId??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQBYTE24) ;
            os << hex << (unsigned short)md->MsgId??(i++??)) ;
  os << "> " ;
  os << endl ;

  os << "      CorrelId        :"  ;
        for(i=0; i < sizeof(MQBYTE24) ;
            os << md->CorrelId??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQBYTE24) ;
            os << hex << (unsigned short)md->CorrelId??(i++??)) ;
  os << "> " ;
  os << endl ;

  os << "      BackoutCount    :"  ;
        os << md->BackoutCount ;
  os << " <"  ;
        os << hex << md->BackoutCount ;
  os << "> " ;
  os << endl ;

  os << "      ReplyToQ        :"  ;
        for(i=0; i < sizeof(MQCHAR48) ;
            os << md->ReplyToQ??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQCHAR48) ;
            os << hex << (unsigned short)md->ReplyToQ??(i++??)) ;
  os << "> " ;
  os << endl ;

  os << "      ReplyToQMgr     :"  ;
        for(i=0; i < sizeof(MQCHAR48) ;
            os << md->ReplyToQMgr??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQCHAR48) ;
            os << hex << (unsigned short)md->ReplyToQMgr??(i++??)) ;
  os << "> " ;
  os << endl ;

  os << "      UserIdentifier  :"  ;
        for(i=0; i < sizeof(MQCHAR12) ;
            os << md->UserIdentifier??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQCHAR12) ;
            os << hex << (unsigned short)md->UserIdentifier??(i++??)) ;
  os << "> " ;
  os << endl ;

  os << "      AccountingToken :"  ;
        for(i=0; i < sizeof(MQBYTE32) ;
            os << md->AccountingToken??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQBYTE32) ;
            os << hex << (unsigned short)md->AccountingToken??(i++??)) ;
  os << "> " ;
  os << endl ;

  os << "      ApplIdentityData:"  ;
        for(i=0; i < sizeof(MQCHAR32) ;
            os << md->ApplIdentityData??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQCHAR32) ;
            os << hex << (unsigned short)md->ApplIdentityData??(i++??)) ;
  os << "> " ;
  os << endl ;

  os << "      PutApplType     :"  ;
        os << md->PutApplType ;
  os << " <"  ;
        os << hex << md->PutApplType ;
  os << "> " ;
  os << endl ;

  os << "      PutApplName     :"  ;
        for(i=0; i < sizeof(MQCHAR28) ;
            os << md->PutApplName??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQCHAR28) ;
            os << hex << (unsigned short)md->PutApplName??(i++??)) ;
  os << "> " ;
  os << endl ;

  os << "      PutDate         :"  ;
        for(i=0; i < sizeof(MQCHAR8) ;
            os << md->PutDate??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQCHAR8) ;
            os << hex << (unsigned short)md->PutDate??(i++??)) ;
  os << "> " ;
  os << endl ;

  os << "      PutTime         :"  ;
        for(i=0; i < sizeof(MQCHAR8) ;
            os << md->PutTime??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQCHAR8) ;
            os << hex << (unsigned short)md->PutTime??(i++??)) ;
  os << "> " ;
  os << endl ;

  os << "      ApplOriginData  :"  ;
        for(i=0; i < sizeof(MQCHAR4) ;
            os << md->ApplOriginData??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQCHAR4) ;
            os << hex << (unsigned short)md->ApplOriginData??(i++??)) ;
  os << "> " ;
  os << endl ;

  return (os) ;

 }


//
// Internal Functions
//

//
// Tracing is controlled by the DEBUG macro.
//
//         If the dynamic usage of DEBUG is set, then the
//            istrace routine determines whether or not data
//            is written to cout (based on the RXMQCRTRACE Rexx var).
//

//
// istrace will obtain the trace setting currently in effect, match
//                       up a substring of the request, and return
//                       TRUE if that trace is active.
//
//         Known settings are :
//
//                              COM   -> Command function
//
//                              GV    -> Obtaining a Rexx Variable
//                              PV    -> Setting   a Rexx Variable
//                              SK    -> Return Code processing
//
//                              INIT  -> Initialisation processing
//                              TERM  -> Deregistration processing
//
//                              *     -> Trace everything!!!
//
//
//  These settings are extracted from the RXMQCRTRACE Rexx Variable
//    (so RXMQCRTRACE = 'CALL' will trace the MQSC operation)
//
//
//

int  istrace ( const char * test )   //Test setting
{
 RXSTRING                varname          ;  // Variable name
 RXSTRING                varvalu          ;  // Variable setting
 unsigned char           varvalc??(600??) ;  // Char version of variable
 unsigned char           varnamc??(600??) ;  // Char version of var name
 SHVBLOCK                sv1              ;  // Var interface CB
 ULONG                   sv1rc            ;  // Var interface RC
 MQLONG                  number           ;  // Char->Integer Conver
 int                     tf      = FALSE  ;  // Return setting (unfound)


 DEBUGS( "Entering istrace "        ) ;
 DEBUGS( "Test is <" << test << ">" ) ;

 if ( strstr(test          ,"*" ) != 0 ) tf = TRUE ;

 strcpy((char *)varnamc,"RXMQCRTRACE")                             ;
 MAKERXSTRING(varname,varnamc,strlen((const char *)varnamc))       ;
 memset(&varvalu,0,sizeof(varvalu))                                ;
 memset(&varvalc,0,sizeof(varvalc))                                ;
 varvalu.strlength= sizeof(varvalc)                                ;
 varvalu.strptr   = varvalc                                        ;
 sv1.shvnext      = 0                                              ;
 memcpy(&sv1.shvname, &varname,sizeof(varname))                    ;
 sv1.shvnamelen  = varname.strlength                               ;
 memcpy(&sv1.shvvalue,&varvalu,sizeof(varvalu))                    ;
 sv1.shvvaluelen = varvalu.strlength                               ;
 sv1.shvcode     = RXSHV_SYFET                                     ;
 sv1.shvret      = 0                                               ;
 sv1rc           = RexxVariablePool(&sv1)                          ;
 if ( (sv1rc == RXSHV_OK) && (varvalu.strlength != 0) )
    {
      if ( strstr((char *)varvalu.strptr,"*" ) != 0 ) tf = TRUE ;
      if ( strstr((char *)varvalu.strptr,test) != 0 ) tf = TRUE ;
    } ;

 DEBUGS( "rc " << sv1rc << " " << varnamc << " ->" << varvalu.strptr
               << "<- " << strlen((const char *)varvalu.strptr)
               << "/"   << varvalu.strlength << " = " << tf ) ;

 DEBUGS( "Leaving istrace " ) ;

 return(tf) ;

} // End of istrace function

//
//  Functions to extract data from Rexx Variables
//
//       get_var   : <varname>   -> RexxString
//       get_varl  : <varname>   -> unsigned long
//       get_var0  : <varname.>0 -> unsigned long
//       get_var1  : <varname.>1 -> RexxString
//
//       where <variable> is the NAME of the Rexx Variable whose
//                           value is to be obtained
//
//             <variable.> indicates that the named variable is
//                           a Rexx Stem variable (so the name
//                           ends in a dot)
//
//       The functions that return data as a RXSTRING require as input
//           the RXSTRING in question; then string pointer part will
//           be updated to contain a pointer to a Rexx-obtained
//           area. This area must be Freemained when usage of the
//           RXString has ended.
//
//       I have deliberately duplicated these functions, so that
//         processing can be readily adapted for subsequent
//         customer extension (although I know this is a bad show!).
//


//
// get_var will obtain a Rexx item as a RXSTRING
//

void get_var (  RXSTRING        name ,  //Var name (without .1)
                RXSTRING      * valu    //Var value
             )
{
 RXSTRING                varname          ;  // Variable name
 RXSTRING                varvalu          ;  // Variable setting
 unsigned char           varnamc??(600??) ;  // Char version of var name
 SHVBLOCK                sv1              ;  // Var interface CB
 ULONG                   sv1rc            ;  // Var interface RC
 MQLONG                  number           ;  // Char->Integer Conver


 DEBUG("GV", "Entering get_var for " << name.strptr ) ;

 sprintf((char *)varnamc,"%s",name.strptr)                         ;
 MAKERXSTRING(varname,varnamc,strlen((const char *)varnamc))       ;
 memset(&varvalu,0,sizeof(varvalu))                                ;
 varvalu.strlength= 0                                              ;
 varvalu.strptr   = 0                                              ;
 sv1.shvnext      = 0                                              ;
 memcpy(&sv1.shvname, &varname,sizeof(varname))                    ;
 sv1.shvnamelen  = varname.strlength                               ;
 memcpy(&sv1.shvvalue,&varvalu,sizeof(varvalu))                    ;
 sv1.shvvaluelen = varvalu.strlength                               ;
 sv1.shvcode     = RXSHV_SYFET                                     ;
 sv1.shvret      = 0                                               ;
 sv1rc           = RexxVariablePool(&sv1)                          ;
 valu->strlength = sv1.shvvalue.strlength                          ;
 valu->strptr    = sv1.shvvalue.strptr                             ;

 DEBUG("GV", "rc " << sv1rc << " " << varnamc << " ->" << valu->strptr
          << "<- " << strlen((const char *)valu->strptr)
          << "/"   << valu->strlength ) ;

 DEBUG("GV", "Leaving get_var " ) ;

 return ;

} // End of get_var function

//
// get_varl will obtain the stem.n of a REXX Stem variable as an
//                                      unsigned long.
//

void get_varl (  RXSTRING    name   ,  //Var name
                 MQLONG    * valu      //Var value
              )
{
 RXSTRING                varname          ;  // Variable name
 RXSTRING                varvalu          ;  // Variable setting
 unsigned char           varvalc??(600??) ;  // Char version of variable
 unsigned char           varnamc??(600??) ;  // Char version of var name
 SHVBLOCK                sv1              ;  // Var interface CB
 ULONG                   sv1rc            ;  // Var interface RC
 MQLONG                  number           ;  // Char->Integer Conver

 DEBUG("GV", "Entering get_varl for " << name.strptr ) ;

 strcpy((char *)varnamc,(const char *)name.strptr)                 ;
 MAKERXSTRING(varname,varnamc,strlen((const char *)varnamc))       ;
 memset(&varvalu,0,sizeof(varvalu))                                ;
 memset(&varvalc,0,sizeof(varvalc))                                ;
 varvalu.strlength= sizeof(varvalc)                                ;
 varvalu.strptr   = varvalc                                        ;
 sv1.shvnext      = 0                                              ;
 memcpy(&sv1.shvname, &varname,sizeof(varname))                    ;
 sv1.shvnamelen  = varname.strlength                               ;
 memcpy(&sv1.shvvalue,&varvalu,sizeof(varvalu))                    ;
 sv1.shvvaluelen = varvalu.strlength                               ;
 sv1.shvcode     = RXSHV_SYFET                                     ;
 sv1.shvret      = 0                                               ;
 sv1rc           = RexxVariablePool(&sv1)                          ;
 number          = atol((const char *)varvalu.strptr)              ;
 if ( (sv1rc == RXSHV_OK) && (varvalu.strlength != 0) )
           *valu = number ;
      else *valu = 0      ;

 DEBUG("GV", "rc " << sv1rc << " " << varnamc << " ->" << varvalu.strptr
          << "<- " << strlen((const char *)varvalu.strptr)
          << "/"   << varvalu.strlength ) ;


 DEBUG("GV", "Leaving get_varl " ) ;

 return ;

} // End of get_varl function

//
// get_var0 will obtain the stem.0 of a REXX Stem variable as an
//                                      unsigned long.
//

void get_var0 (  RXSTRING    name ,  //Var name (without .0)
                 MQLONG    * valu    //Var value
              )
{
 RXSTRING                varname          ;  // Variable name
 RXSTRING                varvalu          ;  // Variable setting
 unsigned char           varvalc??(600??) ;  // Char version of variable
 unsigned char           varnamc??(600??) ;  // Char version of var name
 SHVBLOCK                sv1              ;  // Var interface CB
 ULONG                   sv1rc            ;  // Var interface RC
 MQLONG                  number           ;  // Char->Integer Conver

 DEBUG("GV", "Entering get_var0 for " << name.strptr ) ;

 sprintf((char *)varnamc,"%s0",name.strptr)                        ;
 MAKERXSTRING(varname,varnamc,strlen((const char *)varnamc))       ;
 memset(&varvalu,0,sizeof(varvalu))                                ;
 memset(&varvalc,0,sizeof(varvalc))                                ;
 varvalu.strlength= sizeof(varvalc)                                ;
 varvalu.strptr   = varvalc                                        ;
 sv1.shvnext      = 0                                              ;
 memcpy(&sv1.shvname, &varname,sizeof(varname))                    ;
 sv1.shvnamelen  = varname.strlength                               ;
 memcpy(&sv1.shvvalue,&varvalu,sizeof(varvalu))                    ;
 sv1.shvvaluelen = varvalu.strlength                               ;
 sv1.shvcode     = RXSHV_SYFET                                     ;
 sv1.shvret      = 0                                               ;
 sv1rc           = RexxVariablePool(&sv1)                          ;
 number          = atol((const char *)varvalu.strptr)              ;
 if ( (sv1rc == RXSHV_OK) && (varvalu.strlength != 0) )
           *valu = number ;
      else *valu = 0      ;

 DEBUG("GV", "rc " << sv1rc << " " << varnamc << " ->" << varvalu.strptr
          << "<- " << strlen((const char *)varvalu.strptr)
          << "/"   << varvalu.strlength ) ;


 DEBUG("GV", "Leaving get_var0 " ) ;

 return ;

} // End of get_var0 function

//
// get_var1 will obtain the stem.1 of a REXX Stem variable as a
//                                      RXSTRING
//

void get_var1 (  RXSTRING        name ,  //Var name (without .1)
                 RXSTRING      * valu    //Var value
              )
{
 RXSTRING                varname          ;  // Variable name
 RXSTRING                varvalu          ;  // Variable setting
 unsigned char           varnamc??(600??) ;  // Char version of var name
 SHVBLOCK                sv1              ;  // Var interface CB
 ULONG                   sv1rc            ;  // Var interface RC
 MQLONG                  number           ;  // Char->Integer Conver


 DEBUG("GV", "Entering get_var1 for " << name.strptr ) ;

 sprintf((char *)varnamc,"%s1",name.strptr)                        ;
 MAKERXSTRING(varname,varnamc,strlen((const char *)varnamc))       ;
 memset(&varvalu,0,sizeof(varvalu))                                ;
 sv1.shvnext      = 0                                              ;
 memcpy(&sv1.shvname, &varname,sizeof(varname))                    ;
 sv1.shvnamelen  = varname.strlength                               ;
 memcpy(&sv1.shvvalue,&varvalu,sizeof(varvalu))                    ;
 sv1.shvvaluelen = varvalu.strlength                               ;
 sv1.shvcode     = RXSHV_SYFET                                     ;
 sv1.shvret      = 0                                               ;
 sv1rc           = RexxVariablePool(&sv1)                          ;
 valu->strlength = sv1.shvvalue.strlength                          ;
 valu->strptr    = sv1.shvvalue.strptr                             ;

 DEBUG("GV", "rc " << sv1rc << " " << varnamc << " ->" << valu->strptr
          << "<- " << strlen((const char *)valu->strptr)
          << "/"   << valu->strlength ) ;

 DEBUG("GV", "Leaving get_var1 " ) ;


 return ;

} // End of get_var1 function

//
//  Functions to place data in Rexx Variables
//
//       put_var   : RexxString    -> <varname>
//       put_var0  : unsigned long -> <varname.>0
//       put_var1  : RexxString    -> <varname.>1
//
//       where <variable> is the NAME of the Rexx Variable whose
//                           value is to be set
//
//             <variable.> indicates that the named variable is
//                           a Rexx Stem variable (so the name
//                           ends in a dot)
//
//       Note that, as all Rexx Variables are character strings,
//            the general use of put_var requires that the data
//            has already been converted into a string format if
//            it was of a numeric type.
//
//       I have deliberately duplicated these functions, so that
//         processing can be readily adapted for subsequent
//         customer extension (although I know this is a bad show!).
//
//       I know that I could have used C++ polymorphic matching
//         to use merely a single function name and get the right
//         one with overloading ..... but call me old fashioned, I
//         happen to think that with non-class functions it is better
//         to avoid this and use explicit function names.
//
//

//
// put_var0 will set the stem.0 of a REXX Stem variable as an
//                                   signed long.
//

void put_var0 (  RXSTRING    name ,  //Var name (without .0)
                 long        valu    //Var value
              )
{
 RXSTRING                varname          ;  // Variable name
 RXSTRING                varvalu          ;  // Variable setting
 unsigned char           varvalc??(600??) ;  // Char version of variable
 unsigned char           varnamc??(600??) ;  // Char version of var name
 SHVBLOCK                sv1              ;  // Var interface CB
 ULONG                   sv1rc            ;  // Var interface RC
 MQLONG                  number           ;  // Char->Integer Conver


 DEBUG("PV", "Entering put_var0 for " << name.strptr) ;
 DEBUG("PV", varnamc << " ->" << varvalu.strptr
          << "<- " << strlen((const char *)varvalu.strptr)
          << "/"   << varvalu.strlength ) ;


 sprintf((char *)varvalc,"%lu",valu)                               ;
 sprintf((char *)varnamc,"%s0",name.strptr)                        ;
 MAKERXSTRING(varname,varnamc,strlen((const char *)varnamc))       ;
 MAKERXSTRING(varvalu,varvalc,strlen((const char *)varvalc))       ;
 sv1.shvnext      = 0                                              ;
 memcpy(&sv1.shvname, &varname,sizeof(varname))                    ;
 memcpy(&sv1.shvvalue,&varvalu,sizeof(varvalu))                    ;
 sv1.shvvaluelen = varvalu.strlength                               ;
 sv1.shvcode     = RXSHV_SYSET                                     ;
 sv1.shvret      = 0                                               ;
 sv1rc           = RexxVariablePool(&sv1)                          ;

 DEBUG("PV", "rc " << sv1rc << " " << varnamc << " ->" << varvalu.strptr
          << "<- " << strlen((const char *)varvalu.strptr)
          << "/"   << varvalu.strlength ) ;

 DEBUG("PV", "Leaving put_var0 " ) ;


 return ;

} // End of put_var0 function

//
// put_var1 will set the stem.1 of a REXX Stem variable as a
//                                   RXSTRING
//

void put_var1 (  RXSTRING        name  ,  //Var name (without .1)
                 MQLONG          valul ,  //Var value length
                 void          * valu     //Var value
              )
{
 RXSTRING                varname          ;  // Variable name
 RXSTRING                varvalu          ;  // Variable setting
 unsigned char           varvalc??(600??) ;  // Char version of variable
 unsigned char           varnamc??(600??) ;  // Char version of var name
 SHVBLOCK                sv1              ;  // Var interface CB
 ULONG                   sv1rc            ;  // Var interface RC


 DEBUG("PV", "Entering put_var1 for " << name.strptr) ;
 DEBUG("PV", "Pointer is " << hex << valu << " Len is " << valul ) ;

 sprintf((char *)varnamc,"%s1",name.strptr)                        ;
 MAKERXSTRING(varname,varnamc,strlen((const char *)varnamc))       ;
 MAKERXSTRING(varvalu,valu,valul)                                  ;
 sv1.shvnext      = 0                                              ;
 memcpy(&sv1.shvname, &varname,sizeof(varname))                    ;
 memcpy(&sv1.shvvalue,&varvalu,sizeof(varvalu))                    ;
 sv1.shvvaluelen = valul                                           ;
 sv1.shvcode     = RXSHV_SYSET                                     ;
 sv1.shvret      = 0                                               ;
 sv1rc           = RexxVariablePool(&sv1)                          ;

 DEBUG("PV", "rc " << sv1rc << " " << varnamc << " ->" << varvalu.strptr
          << "<- " << strlen((const char *)varvalu.strptr)
          << "/"   << varvalu.strlength ) ;

 DEBUG("PV", "Leaving put_var1 " ) ;

 return ;

} // End of put_var1 function

//
// put_var will write a Rexx Variable
//
//

void put_var (  RXSTRING name ,  //Var value
                RXSTRING valu    //Var name
             )
{
 SHVBLOCK                sv1              ;  // Var interface CB
 ULONG                   sv1rc            ;  // Var interface RC


 DEBUG("PV", "Entering put_var for " << name.strptr) ;

 sv1.shvnext      = 0                            ;
 memcpy(&sv1.shvname,&name,sizeof(name))         ;
 sv1.shvnamelen  = name.strlength                ;
 memcpy(&sv1.shvvalue,&valu,sizeof(valu))        ;
 sv1.shvvaluelen = valu.strlength                ;
 sv1.shvcode     = RXSHV_SYSET                   ;
 sv1.shvret      = 0                             ;
 sv1rc           = RexxVariablePool(&sv1)        ;

 DEBUG("PV", "rc " << sv1rc << " " << name.strptr << " ->" << valu.strptr
          << "<- " << strlen((const char *)valu.strptr)
          << "/"   << valu.strlength ) ;

 DEBUG("PV", "Leaving put_var " ) ;

 return ;
} // End of put_var function


//
// Functions that manipulate Rexx Stem Variables from the User
//
//           The get_parms_from_stem  function unload the stem variable
//                                    representing the QM and Queues to
//                                    use for the MQSC operation. It is
//                                    a stem variable (whose name ending
//                                    with a dot) whose name is supplied.
//                                    If the .xx part is present, then
//                                    the contents are copied into the
//                                    return parms.
//
//
//

//
// get_parms_from_stem will return the names of the Queue Manager,
//                   Command Queue, and Model ReplyToQ from the Stem
//                   variable.
//
//                     .QM   -> Queue Manager Name
//                     .CQ   -> QM Command Queue
//                                 (def SYSTEM.ADMIN.COMMAND.QUEUE)
//                     .RQ   -> the ReplyToQ Model
//                                 (def SYSTEM.MQSC.REPLY.QUEUE)
//                     .TO   -> Timeout for the MQGET
//                                 (def 5 seconds)
//
//        PS: the defaults are set by the caller!
//

void get_parms_from_stem ( char     * qm     , //QM Name
                           char     * cq     , //Command Queue Name
                           char     * rq     , //ReplyToQ Name
                           MQLONG   * to     , //Timeout value
                           RXSTRING   stem     //name of stem variable
                       )
{
 RXSTRING                varname          ;  // Variable name
 unsigned char           varnamc??(600??) ;  // Char version of variable
 RXSTRING                varvalu          ;  // Variable setting
 unsigned char           varvalc??(600??) ;  // Char version of variable
 SHVBLOCK                sv1              ;  // Var interface CB
 ULONG                   sv1rc            ;  // Var interface RC
 MQLONG                  number           ;  // Char->Integer Conver
 unsigned int            varlen           ;  // Length of Stem.? var


 DEBUG("GV", "Entering get_parms_from_stem " ) ;


 //If the given variable is not a stem. variable, then take the
 //   quicker option of assuming it's the Queue Manager name

 if ( stem.strptr??((strlen((const char *)stem.strptr)-1)??) != '.' )
    {
     memset(qm,0,MQ_Q_MGR_NAME_LENGTH)          ;
     memcpy(qm,
            stem.strptr,
            strlen((const char *)stem.strptr) ) ;
     DEBUG("GV", "Queue Manager " << qm  )      ;

     return ;
    }

 //The given variable is a stem. variable, so get its contents

 sprintf((char *)varnamc,"%sQM",stem.strptr)                       ;
 varlen = strlen((const char *)varnamc)                            ;
 MAKERXSTRING(varname,varnamc,strlen((const char *)varnamc))       ;
 memset(&varvalu,0,sizeof(varvalu))                                ;
 memset(&varvalc,0,sizeof(varvalc))                                ;
 varvalu.strlength= sizeof(varvalc)                                ;
 varvalu.strptr   = varvalc                                        ;
 sv1.shvnext      = 0                                              ;
 memcpy(&sv1.shvname, &varname,sizeof(varname))                    ;
 sv1.shvnamelen  = varname.strlength                               ;
 memcpy(&sv1.shvvalue,&varvalu,sizeof(varvalu))                    ;
 sv1.shvvaluelen = varvalu.strlength                               ;
 sv1.shvcode     = RXSHV_SYFET                                     ;
 sv1.shvret      = 0                                               ;
 sv1rc           = RexxVariablePool(&sv1)                          ;
 if (    (sv1rc                                    == RXSHV_OK)
      && (varvalu.strlength                        != 0       )
      && (memicmp(varnamc, varvalu.strptr, varlen) != 0       ) )
      {
        if (varvalu.strlength > MQ_Q_NAME_LENGTH  )
            varvalu.strlength = MQ_Q_NAME_LENGTH  ;

        memset(qm,0,MQ_Q_NAME_LENGTH)             ;
        memcpy(qm,
               varvalu.strptr,
               strlen((const char *)varvalu.strptr) ) ;
      }

 DEBUG("GV", "rc " << sv1rc << " " << varnamc << " ->" << varvalu.strptr
           << "<- " << strlen((const char *)varvalu.strptr)
           << "/"   << varvalu.strlength ) ;


 sprintf((char *)varnamc,"%sCQ",stem.strptr)                       ;
 varlen = strlen((const char *)varnamc)                            ;
 MAKERXSTRING(varname,varnamc,strlen((const char *)varnamc))       ;
 memset(&varvalu,0,sizeof(varvalu))                                ;
 memset(&varvalc,0,sizeof(varvalc))                                ;
 varvalu.strlength= sizeof(varvalc)                                ;
 varvalu.strptr   = varvalc                                        ;
 sv1.shvnext      = 0                                              ;
 memcpy(&sv1.shvname, &varname,sizeof(varname))                    ;
 sv1.shvnamelen  = varname.strlength                               ;
 memcpy(&sv1.shvvalue,&varvalu,sizeof(varvalu))                    ;
 sv1.shvvaluelen = varvalu.strlength                               ;
 sv1.shvcode     = RXSHV_SYFET                                     ;
 sv1.shvret      = 0                                               ;
 sv1rc           = RexxVariablePool(&sv1)                          ;
 if (    (sv1rc                                    == RXSHV_OK)
      && (varvalu.strlength                        != 0       )
      && (memicmp(varnamc, varvalu.strptr, varlen) != 0       ) )
      {
        if (varvalu.strlength > MQ_Q_NAME_LENGTH  )
            varvalu.strlength = MQ_Q_NAME_LENGTH  ;

        memset(cq,0,MQ_Q_NAME_LENGTH)             ;
        memcpy(cq,
               varvalu.strptr,
               strlen((const char *)varvalu.strptr) ) ;
      }

 DEBUG("GV", "rc " << sv1rc << " " << varnamc << " ->" << varvalu.strptr
           << "<- " << strlen((const char *)varvalu.strptr)
           << "/"   << varvalu.strlength ) ;


 sprintf((char *)varnamc,"%sRQ",stem.strptr)                       ;
 varlen = strlen((const char *)varnamc)                            ;
 MAKERXSTRING(varname,varnamc,strlen((const char *)varnamc))       ;
 memset(&varvalu,0,sizeof(varvalu))                                ;
 memset(&varvalc,0,sizeof(varvalc))                                ;
 varvalu.strlength= sizeof(varvalc)                                ;
 varvalu.strptr   = varvalc                                        ;
 sv1.shvnext      = 0                                              ;
 memcpy(&sv1.shvname, &varname,sizeof(varname))                    ;
 sv1.shvnamelen  = varname.strlength                               ;
 memcpy(&sv1.shvvalue,&varvalu,sizeof(varvalu))                    ;
 sv1.shvvaluelen = varvalu.strlength                               ;
 sv1.shvcode     = RXSHV_SYFET                                     ;
 sv1.shvret      = 0                                               ;
 sv1rc           = RexxVariablePool(&sv1)                          ;
 if (    (sv1rc                                    == RXSHV_OK)
      && (varvalu.strlength                        != 0       )
      && (memicmp(varnamc, varvalu.strptr, varlen) != 0       ) )
      {
        if (varvalu.strlength > MQ_Q_NAME_LENGTH  )
            varvalu.strlength = MQ_Q_NAME_LENGTH  ;

        memset(rq,0,MQ_Q_NAME_LENGTH)             ;
        memcpy(rq,
               varvalu.strptr,
               strlen((const char *)varvalu.strptr) ) ;
      }

 DEBUG("GV", "rc " << sv1rc << " " << varnamc << " ->" << varvalu.strptr
           << "<- " << strlen((const char *)varvalu.strptr)
           << "/"   << varvalu.strlength ) ;

 sprintf((char *)varnamc,"%sTO",stem.strptr)                       ;
 varlen = strlen((const char *)varnamc)                            ;
 MAKERXSTRING(varname,varnamc,strlen((const char *)varnamc))       ;
 memset(&varvalu,0,sizeof(varvalu))                                ;
 memset(&varvalc,0,sizeof(varvalc))                                ;
 varvalu.strlength= sizeof(varvalc)                                ;
 varvalu.strptr   = varvalc                                        ;
 sv1.shvnext      = 0                                              ;
 memcpy(&sv1.shvname, &varname,sizeof(varname))                    ;
 sv1.shvnamelen  = varname.strlength                               ;
 memcpy(&sv1.shvvalue,&varvalu,sizeof(varvalu))                    ;
 sv1.shvvaluelen = varvalu.strlength                               ;
 sv1.shvcode     = RXSHV_SYFET                                     ;
 sv1.shvret      = 0                                               ;
 sv1rc           = RexxVariablePool(&sv1)                          ;
 number          = atol((const char *)varvalu.strptr)              ;
 if (    (sv1rc                                    == RXSHV_OK)
      && (varvalu.strlength                        != 0       )
      && (memicmp(varnamc, varvalu.strptr, varlen) != 0       ) )
      {
        *to = number ;
      }

 DEBUG("GV", "rc " << sv1rc << " " << varnamc << " ->" << varvalu.strptr
           << "<- " << strlen((const char *)varvalu.strptr)
           << "/"   << varvalu.strlength ) ;


 DEBUG("GV", "Leaving get_parms_from_stem " ) ;


 return ;
} // End of get_parms_from_stem function


//
// External Functions, callable from Rexx
//

//
// Initialise the interface - RXMQCRINIT
//
//    This function installs the other functions of the DLL.
//    It MUST be called before any of the other Rexx/MQ functions are
//    available for usage.
//
//    This is done by a RxFuncAdd('RXMQCRINIT','RXMQCR',RXMQCRINIT')
//         call to let Rexx know about the DLL, and then a
//         rcc = RXMQCRINIT() to call the initialisation function.
//
//    The following things are done:
//
//        * Make the RXMQCR functions known to Rexx
//
//
//

LONG _Export _System  rxmqcrinit
                      ( PSZ       afuncname   ,  //Rexx Func name
                        LONG      aargc       ,  //     Number of parms
                        RXSTRING  aargv??(??) ,  //     Parms
                        PSZ       aqname      ,  //     Cur Queue
                        PRXSTRING aretstr     )  //     Return String
 {

 ULONG                    i                ;  // Looper
 RXSTRING                 intname          ;  // Variable name
 RXSTRING                 intvalu          ;  // Variable setting
 struct s_define_mq_ints *pi               ;  // Table pointer
 unsigned char            intvalc??(600??) ;  // Char version of variable
 unsigned char            returnc??(600??) ;  // Return String
 SHVBLOCK                 sv1              ;  // Var interface CB
 ULONG                    sv1rc            ;  // Var interface RC
 ULONG                    rc               ;  // Function interface


 //
 // This is a table of functions to be registered with Rexx
 //

 static char * funcs??(??) = { "RXMQCRC"   ,  "RXMQCRTERM"   } ;

 static unsigned long numfuncs = sizeof(funcs)/sizeof(char *) ;


 DEBUG ("INIT", "Entering rxmqcrinit" ) ;

    // Register the DLL functions

    for (i= 0; i< numfuncs ; i++ )
        {
         rc = RexxRegisterFunctionDll((unsigned char *)funcs??(i??),
                                      (unsigned char *)THISDLL,
                                      (unsigned char *)funcs??(i??)) ;

         DEBUG ("INIT", "Registration of " << funcs??(i??) << " RC " << rc ) ;

        } ;


    // Return a good return Code

    strcpy((char *)returnc,"0 0 0 RXMQCRINIT OK " \
                           "Written by Robert Harris. Version 1.0 " \
                           "Copyright IBM UK Labs Ltd., 1995. " ) ;

    aretstr->strlength  = strlen((const char *)returnc)   ;
    aretstr->strptr     = returnc                         ;

    DEBUG("INIT", "Leaving rxmqcrinit " ) ;
    return 0 ;

 } // End of RXMQCRINIT function


//
// Terminate the interface - RXMQCRTERM
//
//    This function removes access to the functions of the DLL.
//         After calling, no other Rexx/MQ functions are available.
//
//
//

LONG _Export _System  rxmqcrterm
                      ( PSZ       afuncname   ,  //Rexx Func name
                        LONG      aargc       ,  //     Number of parms
                        RXSTRING  aargv??(??) ,  //     Parms
                        PSZ       aqname      ,  //     Cur Queue
                        PRXSTRING aretstr     )  //     Return String
 {

 ULONG                    i                ;  // Looper
 unsigned char            returnc??(600??) ;  // Return String
 ULONG                    rc               ;  // Function interface


 //
 // This is a table of functions to be deregistered from Rexx
 //

 static char * funcs??(??) = { "RXMQCRC"     ,
                               "RXMQCRTERM"  ,  "RXMQCRINIT"   } ;

 static unsigned long numfuncs = sizeof(funcs)/sizeof(char *) ;


 DEBUG ("TERM", "Entering rxmqcrterm" ) ;

 // Deregister the DLL functions

 for (i= 0; i< numfuncs ; i++ )
     {
      rc = RexxDeregisterFunction((unsigned char *)funcs??(i??)) ;

      DEBUG ("TERM", "Deregistration of " << funcs??(i??) << " RC " << rc ) ;

     } ;


 // Return a good return Code

 strcpy((char *)returnc,"0 0 0 RXMQCRTERM OK " \
                        "Written by Robert Harris. " \
                        "Rexx MQSC Functions are no longer available " ) ;

 aretstr->strlength  = strlen((const char *)returnc)   ;
 aretstr->strptr     = returnc                         ;

 DEBUG("TERM", "Leaving rxmqcrterm " ) ;
 return 0 ;

 } // End of RXMQCRTERM function


//
// Action a Command
//
//   Call:   rc = rxmqcrc(parms, input_command, output_response )
//

LONG _Export _System rxmqcrc
                     (PSZ        afuncname   ,  //Rexx Func name
                      LONG       aargc       ,  //     Number of parms
                      PRXSTRING  aargv       ,  //     Parms
                      PSZ        aqname      ,  //     Cur Queue
                      PRXSTRING  aretstr     )  //     Return String
 {

 ULONG                   i                ;  // Looper
 unsigned char           returnc??(600??) ;  // Return String
 LONG                    rc      = 0      ;  // Function Return Code
 MQLONG                  ac      = 0      ;  // MQ instance Return Code
 MQLONG                  mqrc    = 0      ;  // MQ RC
 MQLONG                  mqac    = 0      ;  // MQ AC
 APIRET                  dosrc   = 0      ;  // Dos Getmain/Freemain rc

 RXSTRING                vparm            ;  // Variable Parms
 RXSTRING                vcomm            ;  // Variable Command
 RXSTRING                vresp            ;  // Variable Reply Stem Var

 RXSTRING                comm             ;  // Variable Command
 RXSTRING                resp             ;  // Variable Response Stem


 rahmqo      cmq                          ; //MQ instances - command Queue
 rahmqo      rmq                          ; //             - replyto Queue

 char        qm   ??(MQ_Q_MGR_NAME_LENGTH+1??) ; //QM name
 char        cq   ??(MQ_Q_NAME_LENGTH+1    ??) ; //Q  name - command
 char        rq   ??(MQ_Q_NAME_LENGTH+1    ??) ; //Q  name - replyToq
 MQLONG      to   = 5000                       ; //Timeout for MQ Get

 char        pcfarea  ??(MQCFH_STRUC_LENGTH        +    //PCF area - header
                         MQCFIN_STRUC_LENGTH       +    //         - integer parm
                         MQCFST_STRUC_LENGTH_FIXED +    //         - string parm
                         MAXCOMMLEN+5             ??) ; //         - command

 char      * ptrpcf  = (char   *) pcfarea                 ; //Set PCF area pointers
 MQCFH     * ptrpcf1 = (MQCFH  *) pcfarea                 ;
 MQCFIN    * ptrpcf2 = (MQCFIN *)&pcfarea??(MQCFH_STRUC_LENGTH??)    ;
 MQCFST    * ptrpcf3 = (MQCFST *)&pcfarea??(MQCFH_STRUC_LENGTH+MQCFIN_STRUC_LENGTH??) ;
 char      * ptrpcf4 = (char   *)&pcfarea??(MQCFH_STRUC_LENGTH+MQCFIN_STRUC_LENGTH+MQCFST_STRUC_LENGTH_FIXED??) ;

 MQCFH     * bufpcf1         ; //PCF pointers
 MQCFIN    * bufpcf2i        ; // for the returned
 MQCFST    * bufpcf2s        ; // info
 MQCFIL    * bufpcf2il       ;
 MQCFSL    * bufpcf2sl       ;


 MQLONG      pcflen   = 0    ; //Size of PCF to be sent

 MQLONG      crc      = 0    ; //Return codes
 MQLONG      rrc      = 0    ;

 FULLWORD    recno   =  0    ; //Obtained record number


 MQOD                    od                   ;  //MQ Open Descriptor
 MQLONG                  gopts        = 0     ;  //MQ Get Message options
 MQLONG                  popts        = 0     ;  //MQ Put Message options
 MQMD                    md                   ;  //MQ Message descriptor
 MQGMO                   gmo                  ;  //MQ Get Message options
 MQPMO                   pmo                  ;  //MQ Put Message options
 void                 *  buffer       = 0     ;  //-> Data buffer
 MQLONG                  bufflen      = 10000 ;  //   Data length max
 MQLONG                  buffcurlen   = 0     ;  //   Data length actual

 char                    command??(MAXCOMMLEN+5??)  ; //Padded Command
 MQLONG                  commandlen                 ; // to send to QM


 DEBUG("COM", "Entering rxmqcrc " ) ;

 strcpy((char *)returnc,"-99 0 0 RXMQCRC UNKNOWN FAILURE" ) ;

//
// Check the parms
//

 if ( (rc == 0) && (aargc != 3 ) )
    {
     strcpy((char *)returnc,"-1 0 0 RXMQCRC Bad number of parms" ) ;
     rc   = -1 ;
     mqrc =  0 ;
     mqac =  0 ;
    } ;

if ( (rc == 0) && RXNULLSTRING(aargv??(0??)) )
    {
     strcpy((char *)returnc,"-2 0 0 RXMQCRC Null parms" ) ;
     rc   = -2 ;
     mqrc =  0 ;
     mqac =  0 ;
    } ;

 if ( (rc == 0) && RXZEROLENSTRING(aargv??(0??)) )
    {
     strcpy((char *)returnc,"-3 0 0 RXMQCRC Zero parms" ) ;
     rc   = -3 ;
     mqrc =  0 ;
     mqac =  0 ;
    } ;

 if ( (rc == 0) && RXNULLSTRING(aargv??(1??)) )
    {
     strcpy((char *)returnc,"-4 0 0 RXMQCRC Null command var" ) ;
     rc   = -4 ;
     mqrc =  0 ;
     mqac =  0 ;
    } ;

 if ( (rc == 0) && RXZEROLENSTRING(aargv??(1??)) )
    {
     strcpy((char *)returnc,"-5 0 0 RXMQCRC Zero command var" ) ;
     rc   = -5 ;
     mqrc =  0 ;
     mqac =  0 ;
    } ;

 if ( (rc == 0) && RXNULLSTRING(aargv??(2??)) )
    {
     strcpy((char *)returnc,"-6 0 0 RXMQCRC Null response stem var" ) ;
     rc   = -6 ;
     mqrc =  0 ;
     mqac =  0 ;
    } ;

 if ( (rc == 0) && RXZEROLENSTRING(aargv??(2??)) )
    {
     strcpy((char *)returnc,"-7 0 0 RXMQCRC Zero length response stem var" ) ;
     rc   = -7 ;
     mqrc =  0 ;
     mqac =  0 ;
    } ;


//
// Now the parms are correct, get them
//

 if (rc == 0)
    {
      memcpy(&vparm,&aargv??(0??),sizeof(vparm)) ; //QM Name (or stem.)
      memcpy(&vcomm,&aargv??(1??),sizeof(vcomm)) ; //Command to issue
      memcpy(&vresp,&aargv??(2??),sizeof(vresp)) ; //Return stem.

      DEBUG("COM",  "vparm   " << vparm.strptr   ) ;
      DEBUG("COM",  "vcomm   " << vcomm.strptr   ) ;
      DEBUG("COM",  "vresp   " << vresp.strptr   ) ;


      memset(qm,0,MQ_Q_MGR_NAME_LENGTH+1     ) ; //Clear data areas
      memset(cq,0,MQ_Q_NAME_LENGTH+1         ) ;
      memset(rq,0,MQ_Q_NAME_LENGTH+1         ) ;

      strcpy(cq,"SYSTEM.ADMIN.COMMAND.QUEUE" ) ; // and set defaults
      strcpy(rq,"SYSTEM.MQSC.REPLY.QUEUE"    ) ;

      get_parms_from_stem(qm,cq,rq,&to,vparm)  ; //Get the supplied parms
      get_var (vresp, &resp  )                 ; //Get the return stem name

      DEBUG("COM", "QM is " << qm   )          ;
      DEBUG("COM", "CQ is " << rq   )          ;
      DEBUG("COM", "RQ is " << cq   )          ;
      DEBUG("COM", "Command is <" << vcomm.strptr << ">" )  ;

    }


 //If no command, then nothing to do

 if ( (rc == 0 ) && ( vcomm.strlength == 0 ) )
    {
      sprintf((char *)returnc,"-8 0 0 RXMQCRC No command supplied" ) ;
      rc   = -8  ;
      mqrc =  0  ;
      mqac =  0  ;
    } ;

 //If command is too big, error

 if ( (rc == 0 ) && ( vcomm.strlength > MAXCOMMLEN ) )
    {
      sprintf((char *)returnc,"-9 0 0 RXMQCRC Too big a Command supplied" ) ;
      rc   = -9  ;
      mqrc =  0  ;
      mqac =  0  ;
    } ;

//Ensure Command is a multiple of 4 bytes long (PCF String requirement)

if ( rc == 0 )
   {
    memset(command,0,sizeof(command)) ;
    memcpy(command,vcomm.strptr,vcomm.strlength) ;
    switch ( vcomm.strlength % 4 )
     {
       case 1  : strcat(command,"   ") ; break ;
       case 2  : strcat(command,"  " ) ; break ;
       case 3  : strcat(command," "  ) ; break ;
       default : break                         ;
     }
    commandlen = strlen(command) ;
   }



//Set the Put options (for the Command) and Get options (for the reply)

 if ( rc == 0 )                      // Connect to QM
    {
      gopts  = MQGMO_NO_SYNCPOINT         +
               MQGMO_WAIT                 +
               MQGMO_ACCEPT_TRUNCATED_MSG +
               MQGMO_FAIL_IF_QUIESCING    ;

      popts  = MQPMO_NO_SYNCPOINT         +
               MQPMO_DEFAULT_CONTEXT      + //Needed to bluff PCF Security!
               MQPMO_FAIL_IF_QUIESCING    ;
    } ;


//
// Now GETMAIN the buffer to receive the ReplyToQ records
//

 if (( rc == 0) && ( bufflen != 0 ) )
    {
     DEBUG("COM", "Doing DOS Getmain for " << bufflen << " bytes ") ;
     dosrc = DosAllocMem(&buffer,bufflen,PAG_READ|PAG_WRITE|PAG_COMMIT) ;
     DEBUG("COM","Dos Getmain RC " << dosrc << " PTR is "<<hex<< buffer);

     bufpcf1 = (MQCFH *) buffer    ; //Set root for returned PCFs

     if ( dosrc != 0 )
        {
         sprintf((char *)returnc,"-10 0 0 RXMQCRC " \
                                 "DosAllocMem failure, RC(%lu) " ,
                                  dosrc ) ;
         rc   = -10 ;
         mqrc =  0  ;
         mqac =  0  ;
        } ;
    }

 //The return stem variable is initially set to no info

 if ( rc == 0 )
    {
      put_var0(vresp,0) ;
    }

 //Connect to the QM, open the Command Queue, and create the ReplyToQ

 if ( rc == 0 )                      // Connect to QM
    {
      DEBUG ( "COM", "Connecting to QM " << qm ) ;
      ac   = cmq.conn(qm)    ;
      mqrc = cmq.lastrc      ;
      mqac = cmq.lastrc      ;
      DEBUG( "COM", "RC=" << rc ) ;

      if ( ac == 0 ) rmq.qmh = cmq.qmh ;
      if ( ac != 0 )
         {
           rc = ac ;
           sprintf((char *)returnc,"%lu %lu %lu RXMQCRC MQCONN to QM %s %s",
                             ac, mqrc, mqac, qm,
                             ( (mqrc == MQCC_OK      ) ? "OK"      :
                               (mqrc == MQCC_WARNING ) ? "WARNING" :
                               (mqrc == MQCC_FAILED  ) ? "FAILED"  :
                                                         "UNKNOWN" ) ) ;
         }
    }

if ( rc == 0 )                      // Open the command Q for Put access
    {
      DEBUG ( "COM", "Opening Command Q " << cq << " for Output " ) ;
      ac   = cmq.open(cq,(MQOO_OUTPUT+MQOO_FAIL_IF_QUIESCING))   ;
      mqrc = cmq.lastrc     ;
      mqac = cmq.lastrc     ;
      DEBUG ( "COM", "RC=" << rc ) ;

      if ( ac != 0 )
         {
           rc = ac ;
           sprintf((char *)returnc,"%lu %lu %lu RXMQCRC MQOPEN for Command Queue %s %s",
                             ac, mqrc, mqac, cq,
                             ( (mqrc == MQCC_OK      ) ? "OK"      :
                               (mqrc == MQCC_WARNING ) ? "WARNING" :
                               (mqrc == MQCC_FAILED  ) ? "FAILED"  :
                                                         "UNKNOWN" ) ) ;
         }
    }

 if ( rc == 0 )                      // Open/Create the ReplyToQ for Get access
    {

      memset(&od,0,sizeof(od))                             ;
      memcpy(od.StrucId,MQOD_STRUC_ID,sizeof(od.StrucId))  ;
      od.Version        = MQOD_VERSION_1                   ;
      od.ObjectType     = MQOT_Q                           ;
      strcpy(od.ObjectName,rq)                             ;
      strcpy(od.DynamicQName,"RXMQCR.*")                   ;

      DEBUG( "COM","Opening Model ReplyToQ " << od.ObjectName << " for Destructive access " ) ;

      ac   = rmq.open(&od,(MQOO_INPUT_EXCLUSIVE+MQOO_SAVE_ALL_CONTEXT+MQOO_FAIL_IF_QUIESCING)) ;
      mqrc = rmq.lastrc     ;
      mqac = rmq.lastrc     ;
      DEBUG("COM", "RC=" << rc ) ;

      if ( ac != 0 )
         {
           rc = ac ;
           sprintf((char *)returnc,"%lu %lu %lu RXMQCRC MQOPEN for Model ReplyToQ Queue %s %s",
                             ac, mqrc, mqac, rq,
                             ( (mqrc == MQCC_OK      ) ? "OK"      :
                               (mqrc == MQCC_WARNING ) ? "WARNING" :
                               (mqrc == MQCC_FAILED  ) ? "FAILED"  :
                                                         "UNKNOWN" ) ) ;
         }

      if ( ac == 0 )
         {
          memset(&rq,0,sizeof(rq))                        ;
          strncpy(rq,od.ObjectName,sizeof(od.ObjectName)) ;
          DEBUG("COM", "The ReplyToQ name is " << rq ) ;
         }
    }

 //Build the PCF ESCAPE command

 if ( rc == 0 )                      // Now build the PCF command
    {
      DEBUG("COM", "Now building the PCF area") ;
      memset(pcfarea,0,sizeof(pcfarea))             ;
      ptrpcf1->Type           = MQCFT_COMMAND       ; // Header - 2 parms
      ptrpcf1->StrucLength    = MQCFH_STRUC_LENGTH  ;
      ptrpcf1->Version        = MQCFH_VERSION_1     ;
      ptrpcf1->Command        = MQCMD_ESCAPE        ;
      ptrpcf1->MsgSeqNumber   = 1                   ;
      ptrpcf1->Control        = MQCFC_LAST          ;
      ptrpcf1->ParameterCount = 2                   ;
      ptrpcf2->Type           = MQCFT_INTEGER       ; // Parm1 is ESCAPE
      ptrpcf2->StrucLength    = MQCFIN_STRUC_LENGTH ;
      ptrpcf2->Parameter      = MQIACF_ESCAPE_TYPE  ;
      ptrpcf2->Value          = MQET_MQSC           ;
      ptrpcf3->Type           = MQCFT_STRING        ; // Parm2 is ESCAPE data
      ptrpcf3->StrucLength    = MQCFST_STRUC_LENGTH_FIXED + commandlen  ;
      ptrpcf3->Parameter      = MQCACF_ESCAPE_TEXT  ;
      ptrpcf3->CodedCharSetId = MQCCSI_DEFAULT      ;
      ptrpcf3->StringLength   = commandlen          ;
      memcpy(ptrpcf4,command,commandlen)            ;

      pcflen =   MQCFH_STRUC_LENGTH
               + MQCFIN_STRUC_LENGTH
               + MQCFST_STRUC_LENGTH_FIXED
               + commandlen                         ;


    }

// Now write the Command to the relevant Command Queue

 if ( rc == 0)                       //Now write to the Queue
    {
      DEBUG("COM", "Now building the PMD") ;
      memset(&pmo,0,sizeof(pmo))  ;               //Set Put Message Options
      memcpy(&pmo.StrucId,MQPMO_STRUC_ID,sizeof(pmo.StrucId))          ;
      pmo.Version        = MQPMO_VERSION_1                             ;
      pmo.Options        = popts                                       ;

      memset(&cmq.pmd,0,sizeof(cmq.pmd))  ;     //Set Put Message Desc
      memcpy(&cmq.pmd.StrucId,MQMD_STRUC_ID,sizeof(cmq.pmd.StrucId))   ;
      memcpy(&cmq.pmd.Format ,MQFMT_ADMIN  ,sizeof(cmq.pmd.Format ))   ;
      cmq.pmd.Version        = MQMD_VERSION_1                          ;
      cmq.pmd.Report         = MQRO_NONE                               ;
      cmq.pmd.MsgType        = MQMT_REQUEST                            ;
      cmq.pmd.Persistence    = MQPER_NOT_PERSISTENT                    ;
      cmq.pmd.Expiry         = MQEI_UNLIMITED                          ;
      cmq.pmd.Feedback       = MQFB_NONE                               ;
      cmq.pmd.Encoding       = MQENC_NATIVE                            ;
      cmq.pmd.CodedCharSetId = MQCCSI_Q_MGR                            ;
      cmq.pmd.Priority       = MQPRI_PRIORITY_AS_Q_DEF                 ;
      memcpy(&cmq.pmd.ReplyToQ,rq,sizeof(cmq.pmd.ReplyToQ))            ;

      DEBUG("COM", "Now issuing the MQPUT to the Command Queue") ;
      crc = cmq.put( &pmo, pcflen, (char *)ptrpcf1 )        ;
      mqrc = cmq.lastrc     ;
      mqac = cmq.lastrc     ;
      DEBUG ( "COM", "RC=" << rc ) ;

      if ( crc != 0 )
         {
           rc = crc ;
           sprintf((char *)returnc,"%lu %lu %lu RXMQCRC MQPUT for Command Queue %s %s",
                             crc, mqrc, mqac, cq,
                             ( (mqrc == MQCC_OK      ) ? "OK"      :
                               (mqrc == MQCC_WARNING ) ? "WARNING" :
                               (mqrc == MQCC_FAILED  ) ? "FAILED"  :
                                                         "UNKNOWN" ) ) ;
         }


    }

 //Read the ReplyToQ, and display the contained messages

 if ( rc == 0 )                      // Read Q and Display
    {
      DEBUG("COM", "Now starting to obtain the ReplyToQ messages") ;
      for ( recno = 1 ; ; recno++ )
          {
            MQLONG   i      = 0   ;
            char     exiter = 'N' ;
            char   * p      = 0   ;
            MQLONG   mrc    = 0   ;
            MQLONG   frc    = 0   ;
            FULLWORD reclen = 0   ;

            bufpcf1   = (MQCFH  *) buffer                         ; //Point to
            bufpcf2i  = (MQCFIN *)&((char *)buffer)??(MQCFH_STRUC_LENGTH??) ; //Buffer
            bufpcf2s  = (MQCFST *)&((char *)buffer)??(MQCFH_STRUC_LENGTH??) ; //Structures
            bufpcf2il = (MQCFIL *)&((char *)buffer)??(MQCFH_STRUC_LENGTH??) ;
            bufpcf2sl = (MQCFSL *)&((char *)buffer)??(MQCFH_STRUC_LENGTH??) ;

            memset(&gmo,0,sizeof(gmo))                               ; //Build the
            memcpy(&gmo.StrucId,MQGMO_STRUC_ID,sizeof(gmo.StrucId))  ; //GMO, allow
            gmo.Version        = MQGMO_VERSION_1                     ; //? seconds
            gmo.Options        = gopts                               ; //for the
            if ( to != 0 ) gmo.WaitInterval = to                     ; //reply
                 else      gmo.WaitInterval = MQWI_UNLIMITED         ;

            memset(&rmq.gmd,0,sizeof(rmq.gmd))                              ;
            memcpy(&rmq.gmd.StrucId,MQMD_STRUC_ID,sizeof(rmq.gmd.StrucId))  ;
            rmq.gmd.Version        = MQMD_VERSION_1                         ;

            memset(buffer,0,bufflen)                                   ;
            reclen   = 0                                                    ;
            DEBUG("COM", "Issuing a MQGET to the ReplyToQ " << rq ) ;
            rrc = rmq.get ( &gmo, bufflen, (char *)buffer, &reclen )           ;
            DEBUG("COM", "GET RC " << mqrc << " AC " << mqac << " Datalen " << reclen << " " ) ;


            switch ( rrc ) //Print obtained message
             {
               case MQRC_NONE      :
                    rc   = 0       ;

                    DEBUG( "COM", "Message "   << recno
                         << " = MQ Message " << bufpcf1->MsgSeqNumber
                         << ((bufpcf1->Control == MQCFC_LAST) ? " (last)" : " (more)")
                         << " has MQ RC "        << bufpcf1->Reason
                         << ", "                 << bufpcf1->ParameterCount
                         << " parms" ) ;

                    if ( bufpcf1->ParameterCount != 0 ) //Do not attempt to ignore null messages
                       {                                // when printing the results
                        MQLONG   parms           ;
                        int      i               ;
                        int      j               ;
                        MQLONG   parmtype        ;
                        MQLONG   parmsize        ;
                        MQLONG   iparm           ;
                        MQLONG   ival            ;
                        MQCFIL * ip              ;
                        MQCFSL * sp              ;
                        char     resvar??(100??) ;
                        RXSTRING resvarn         ;
                        RXSTRING mess            ;


                        parms = bufpcf1->ParameterCount ;

                        for (i=0 ; i < parms ; i++ )
                            {
                             parmtype = bufpcf2i->Type        ; //All types share a
                             parmsize = bufpcf2i->StrucLength ; // common prefix
                             switch ( parmtype )
                              {
                                case MQCFT_INTEGER  : //Convert to attr=value
                                          DEBUG("COM", " Integer Parm " << bufpcf2i->Parameter
                                                       << " Value "        << bufpcf2i->Value ) ;
                                     break        ;

                                case MQCFT_INTEGER_LIST : //Convert to attr=value pairs
                                     if ( bufpcf2il->Count != 0 )
                                        {
                                          DEBUG("COM", " Integer parm " << bufpcf2il->Parameter
                                               << " Values "  ) ;
                                          for ( j=0 ; j < bufpcf2il->Count ; j++ )
                                              {
                                                ip = bufpcf2il + (sizeof(MQLONG) * j) ;
                                                DEBUG("COM", ip->Values ) ;
                                              } ;
                                        } ;
                                     break        ;


                                case MQCFT_STRING : //Print in the Stem variable

                                     put_var0(vresp,recno)                     ;
                                     sprintf(resvar,"%s%d",resp.strptr,recno)  ;
                                     resvarn.strlength = strlen(resvar)        ;
                                     resvarn.strptr    = (unsigned char *)resvar        ;
                                     mess.strlength =  bufpcf2s->StringLength  ;
                                     mess.strptr    = (unsigned char *)&(bufpcf2s->String)      ;
                                     put_var (resvarn,mess)                    ;
                                     break                                     ;

                                case MQCFT_STRING_LIST : //Print Strings in the Stem Variable
                                     if ( bufpcf2sl->Count != 0 )
                                        {
                                         for ( j=0 ; j < bufpcf2sl->Count ; j++ )
                                             {
                                               sp = bufpcf2sl + ( (bufpcf2sl->StringLength) * j)   ;
                                               put_var0(vresp,recno)                      ;
                                               sprintf(resvar,"%s%d",resp.strptr,recno)  ;
                                               resvarn.strlength = strlen(resvar)        ;
                                               resvarn.strptr    = (unsigned char *)resvar        ;
                                               mess.strlength =  bufpcf2s->StringLength   ;
                                               mess.strptr    = (unsigned char *)&(sp->Strings)            ;
                                               put_var (resvarn,mess)                      ;
                                               break                                      ;
                                             } ;
                                        } ;

                                     break        ;
                                default :
                                     break ;

                              } ;

                             bufpcf2i  = (MQCFIN *)( (char *)bufpcf2i  + parmsize ) ; //Bump
                             bufpcf2s  = (MQCFST *)( (char *)bufpcf2s  + parmsize ) ; // up
                             bufpcf2il = (MQCFIL *)( (char *)bufpcf2il + parmsize ) ; // structure
                             bufpcf2sl = (MQCFSL *)( (char *)bufpcf2sl + parmsize ) ; // pointers

                            } ; //End of Parameter Printing loop

                       } ; //End of Parameter Listing

                    if ( bufpcf1->Control == MQCFC_LAST) exiter = 'L' ; //Last means just that!
                    break          ;

               default             :
                  rc   = rrc            ;
                  mqrc = rmq.lastrc     ;
                  mqac = rmq.lastrc     ;
                  DEBUG("COM", "MQGET on ReplyToQ error RC=" << rrc <<  " on message " << recno ) ;
                  sprintf((char *)returnc,"%lu %lu %lu RXMQCRC MQGET on ReplyToQ %s %s",
                                  ac, mqrc, mqac, rq,
                                  ( (mqrc == MQCC_OK      ) ? "OK"      :
                                    (mqrc == MQCC_WARNING ) ? "WARNING" :
                                    (mqrc == MQCC_FAILED  ) ? "FAILED"  :
                                                              "UNKNOWN" ) ) ;

                    exiter = 'G'  ;
                    rc  = rrc     ;
                    break         ;

             }

              if ( exiter != 'N' ) break ; //Get next message
                                           //unless error/EOQ occured

          } // End of Message Processing FOR loop


    } // End of Queue/Queue Processing Block

  DEBUG("COM", "All ReplyToQ records obtained" ) ;
  if ( rc == 0 ) sprintf((char *)returnc, "0 0 0 RXMQCRC OK"  ) ;

//End of processing

 DEBUG("COM", "Closing the ReplyToQ " ) ; //Close/Delete the ReplyToQ
 ac   = rmq.close(MQCO_NONE)  ;
 mqrc = rmq.lastrc            ;
 mqac = rmq.lastrc            ;
 if ( ac != 0 ) rc = ac       ;
 DEBUG("COM", "RC=" << rc )   ;

 DEBUG("COM", "Closing the command Q " ) ; //Close the Command Queue
 ac   = cmq.close(MQCO_NONE)  ;
 mqrc = cmq.lastrc            ;
 mqac = cmq.lastrc            ;
 if ( ac != 0 ) rc = ac       ;
 DEBUG("COM", "RC=" << rc )   ;

 DEBUG("COM", "Disconnecting from Command QM " ) ; //Disconnect from QM
 ac   = cmq.disc()            ;
 mqrc = cmq.lastrc            ;
 mqac = cmq.lastrc            ;
 if ( ac != 0 ) rc = ac       ;
 DEBUG("COM", "RC=" << rc )   ;


//
// Now FREEMAIN the buffer if it was allocated
//

 if (buffer != 0)
    {
     DEBUG("COM", "Dos Freemaining " ) ;
     dosrc = DosFreeMem(buffer) ;
     DEBUG("COM", "Dos Freemain RC " << dosrc ) ;
     buffer = 0 ;
    }

//
// Set the function return string
//


 aretstr->strlength  = strlen((const char *)returnc)   ;
 aretstr->strptr     = returnc                         ;

//
// Return to caller
//

 DEBUG("COM", "Leaving rxmqcrc " ) ;

 return 0 ;

 } // End of RXMQCRC function


